# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
Module contains ``logger_decorator`` function.

``logger_decorator`` is used for decorating individual Modin functions.
"""

from functools import wraps

from modin.config import LogMode
from .config import get_logger


def logger_decorator(modin_layer: str, function_name: str, log_level: str):
    """
    Log Decorator used on specific internal Modin functions.

    Parameters
    ----------
    modin_layer : str
        Specified by the logger (PANDAS-API).
    function_name : str
        The name of the function the decorator is being applied to.
    log_level : str
        The log level (logging.INFO, logging.DEBUG, logging.WARNING, etc.).

    Returns
    -------
    func
        A decorator function.
    """

    def decorator(f):
        """Decorate function to add logs to Modin API function."""

        @wraps(f)
        def run_and_log(*args, **kwargs):
            """
            Compute function with logging if Modin logging is enabled.

            Parameters
            ----------
            *args : tuple
                The function arguments.
            **kwargs : dict
                The function keyword arguments.

            Returns
            -------
            Any
            """
            if LogMode.get() == "disable":
                return f(*args, **kwargs)

            logger = get_logger()
            try:
                logger_level = getattr(logger, log_level.lower())
            except AttributeError:
                raise AttributeError(f"Invalid log_level: {log_level}")

            logger_level(f"START::{modin_layer.upper()}::{function_name}")
            result = f(*args, **kwargs)
            logger_level(f"STOP::{modin_layer.upper()}::{function_name}")
            return result

        return run_and_log

    return decorator



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
Module contains class ``PyarrowOnRayDataframe``.

``PyarrowOnRayDataframe`` is a dataframe class with PyArrow storage format and Ray engine.
"""

import pandas
from pandas.core.dtypes.cast import find_common_type

from ..partitioning.partition_manager import PyarrowOnRayDataframePartitionManager
from modin.core.dataframe.pandas.dataframe.dataframe import PandasDataframe

import ray


class PyarrowOnRayDataframe(PandasDataframe):
    """
    Class for dataframes with PyArrow storage format and Ray engine.

    ``PyarrowOnRayDataframe`` implements interfaces specific for PyArrow and Ray,
    other functionality is inherited from the ``PandasDataframe`` class.

    Parameters
    ----------
    partitions : np.ndarray
        A 2D NumPy array of partitions.
    index : sequence
        The index for the dataframe. Converted to a ``pandas.Index``.
    columns : sequence
        The columns object for the dataframe. Converted to a ``pandas.Index``.
    row_lengths : list, optional
        The length of each partition in the rows. The "height" of
        each of the block partitions. Is computed if not provided.
    column_widths : list, optional
        The width of each partition in the columns. The "width" of
        each of the block partitions. Is computed if not provided.
    dtypes : pandas.Series, optional
        The data types for the dataframe columns.
    """

    _partition_mgr_cls = PyarrowOnRayDataframePartitionManager

    def synchronize_labels(self, axis=None):
        """
        Synchronize labels by applying the index object (Index or Columns) to the partitions lazily.

        Parameters
        ----------
        axis : {0, 1}, optional
            Parameter is deprecated and affects nothing.
        """
        self._filter_empties()

    @classmethod
    def combine_dtypes(cls, list_of_dtypes, column_names):
        """
        Get common for all partitions dtype for each of the columns.

        Parameters
        ----------
        list_of_dtypes : array-like
            Array with references to the partitions dtypes objects.
        column_names : array-like or pandas.Index
            Columns names to use for resulting Series.

        Returns
        -------
        pandas.Series
            pandas.Series where index is columns names and values are
            columns dtypes.

        Notes
        -----
        The reported dtypes from differing rows can be different based
        on the inference in the limited data seen by each worker. We
        use pandas to compute the exact dtype over the whole column for
        each column.
        """
        dtypes = (
            pandas.concat(ray.get(list_of_dtypes), axis=1)
            .apply(lambda row: find_common_type(row.values), axis=1)
            .squeeze(axis=0)
        )
        dtypes.index = column_names
        return dtypes

    def to_pandas(self):
        """
        Convert frame object to a ``pandas.DataFrame``.

        Returns
        -------
        pandas.DataFrame
        """
        df = super(PyarrowOnRayDataframe, self).to_pandas()
        df.index = self.index
        df.columns = self.columns
        return df



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import pytest
import sys
import subprocess


@pytest.mark.parametrize(
    "import_strategy",
    [
        pytest.param(
            """
import modin.config as cfg
cfg.Engine.put('Native') # 'omniscidbe'/'dbe' would be imported with dlopen flags first time
cfg.StorageFormat.put('Omnisci')
cfg.IsExperimental.put(True)
import modin.pandas as pd
""",
            id="config_omnisci_first-import_modin_second",
        ),
        pytest.param(
            """
import modin.pandas as pd
import modin.config as cfg
cfg.Engine.put('Native')
cfg.StorageFormat.put('Omnisci')
cfg.IsExperimental.put(True)
""",
            id="import_modin_first-config_omnisci_second",
        ),
    ],
)
@pytest.mark.parametrize("has_other_engines", [True, False])
def test_omnisci_import(import_strategy, has_other_engines):
    """
    Test import of OmniSci engine.

    The import of PyDBEngine requires to set special dlopen flags which make it then
    incompatible to import some other libraries further (like ``pyarrow.gandiva``).
    This test verifies that it's not the case when a user naturally imports Modin
    with OmniSci engine.

    Parameters
    ----------
    import_strategy : str
        There are several scenarios of how a user can import Modin with OmniSci engine:
        configure Modin first to use OmniSci engine and then import ``modin.pandas`` or vice versa.
        This parameters holds a python code, implementing one of these scenarios.
    has_other_engines : bool
        The problem with import may appear depending on whether other engines are
        installed. This parameter indicates whether to remove modules for
        non-omnisci engines before the test.

    Notes
    -----
    The failed import flow may cause segfault, which causes to crash the pytest itself.
    This makes us to run the test in a separate process and check its exit-code to
    decide the success of the test.
    """

    remove_other_engines = """
import sys
sys.modules['ray'] = None
sys.modules['dask'] = None
"""

    if not has_other_engines:
        import_strategy = f"{remove_other_engines}\n{import_strategy}"

    res = subprocess.run(
        [sys.executable, "-c", import_strategy],
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    if res.returncode != 0:
        pytest.fail(str(res.stderr))


@pytest.mark.parametrize(
    "import_strategy, expected_to_fail",
    [
        pytest.param(
            """
from modin.experimental.core.execution.native.implementations.omnisci_on_native.utils import PyDbEngine
import pyarrow.gandiva
""",
            True,
            id="import_pydbe_first-pyarrow_gandiva_second",
        ),
        pytest.param(
            """
import pyarrow.gandiva
from modin.experimental.core.execution.native.implementations.omnisci_on_native.utils import PyDbEngine
""",
            False,
            id="import_pyarrow_gandiva_first-pydbe_second",
        ),
    ],
)
def test_omnisci_compatibility_with_pyarrow_gandiva(import_strategy, expected_to_fail):
    """
    Test the current status of compatibility of PyDbEngine and pyarrow.gandiva packages.

    If this test appears to fail, it means that these packages are now compatible/incopmatible,
    if it's so, please post the actual compatibility status to the issue:
    https://github.com/modin-project/modin/issues/3865
    And then inverse `expected_to_fail` parameter for the scenario that has changed its behavior.

    Parameters
    ----------
    import_strategy : str
        There are several scenarios of how a user can import PyDbEngine and pyarrow.gandiva.
        This parameters holds a python code, implementing one of the scenarios.
    expected_to_fail : bool
        Indicates the estimated compatibility status for the specified `import_strategy`.
        True - the strategy expected to fail, False - the strategy expected to pass.
        Note: we can't use built-in ``pytest.marks.xfail`` as we need to check that the
        expected failure was caused by LLVM error.
    """
    res = subprocess.run(
        [sys.executable, "-c", import_strategy],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    if expected_to_fail:
        assert (
            res.returncode != 0
        ), "PyDbEngine and pyarrow.gandiva are now compatible! Please check the test's doc-string for further instructions."
    else:
        assert (
            res.returncode == 0
        ), "PyDbEngine and pyarrow.gandiva are now incompatible! Please check the test's doc-string for further instructions."

    if res.returncode != 0:
        error_msg = res.stderr.decode("utf-8")
        assert (
            error_msg.find("LLVM ERROR") != -1
        ), f"Expected to fail because of LLVM error, but failed because of:\n{error_msg}"



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
Module contains useful enums for Modin.

Axis is an enum that represents the `axis` argument for dataframe operations.
JoinType is an enum that represents the `join_type` or `how` argument for the join algebra operator.
"""

from enum import Enum


class Axis(Enum):  # noqa: PR01
    """
    An enum that represents the `axis` argument provided to the algebra operators.

    The enum has 3 values - ROW_WISE to represent the row axis, COL_WISE to represent the
    column axis, and CELL_WISE to represent no axis. ROW_WISE operations iterate over the rows
    COL_WISE operations over the columns, and CELL_WISE operations over any of the partitioning
    schemes that are supported in Modin (row-wise, column-wise, or block-wise).
    """

    ROW_WISE = 0
    COL_WISE = 1
    CELL_WISE = None


class JoinType(Enum):  # noqa: PR01
    """
    An enum that represents the `join_type` argument provided to the algebra operators.

    The enum has 4 values - INNER to represent inner joins, LEFT to represent left joins, RIGHT to
    represent right joins, and OUTER to represent outer joins.
    """

    INNER = "inner"
    LEFT = "left"
    RIGHT = "right"
    OUTER = "outer"



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""Module holds implementation of ``BaseIO`` using cuDF."""

from modin.core.io import BaseIO
from modin.core.storage_formats.cudf.query_compiler import cuDFQueryCompiler
from modin.core.storage_formats.cudf.parser import cuDFCSVParser
from modin.core.execution.ray.common import RayTask
from ..dataframe import cuDFOnRayDataframe
from ..partitioning import (
    cuDFOnRayDataframePartition,
    cuDFOnRayDataframePartitionManager,
)
from .text import cuDFCSVDispatcher


class cuDFOnRayIO(BaseIO):
    """The class implements ``BaseIO`` class using cuDF-entities."""

    frame_cls = cuDFOnRayDataframe
    query_compiler_cls = cuDFQueryCompiler

    build_args = dict(
        frame_partition_cls=cuDFOnRayDataframePartition,
        query_compiler_cls=cuDFQueryCompiler,
        frame_cls=cuDFOnRayDataframe,
        frame_partition_mgr_cls=cuDFOnRayDataframePartitionManager,
    )

    read_csv = type("", (RayTask, cuDFCSVParser, cuDFCSVDispatcher), build_args).read



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import pandas
import pytest

from modin.pandas.test.utils import (
    test_data_values,
    create_test_dfs,
    df_equals,
)
from modin.config import NPartitions

NPartitions.put(4)


@pytest.mark.parametrize("axis", [0, 1])
@pytest.mark.parametrize("item_length", [0, 1, 2])
@pytest.mark.parametrize("loc", ["first", "first + 1", "middle", "penult", "last"])
@pytest.mark.parametrize("replace", [True, False])
def test_insert_item(axis, item_length, loc, replace):
    data = test_data_values[0]

    def post_fn(df):
        return (
            (df.iloc[:, :-item_length], df.iloc[:, -item_length:])
            if axis
            else (df.iloc[:-item_length, :], df.iloc[-item_length:, :])
        )

    def get_loc(frame, loc):
        locs_dict = {
            "first": 0,
            "first + 1": 1,
            "middle": len(frame.axes[axis]) // 2,
            "penult": len(frame.axes[axis]) - 1,
            "last": len(frame.axes[axis]),
        }
        return locs_dict[loc]

    def get_reference(df, value, loc):
        if axis == 0:
            first_mask = df.iloc[:loc]
            if replace:
                loc += 1
            second_mask = df.iloc[loc:]
        else:
            first_mask = df.iloc[:, :loc]
            if replace:
                loc += 1
            second_mask = df.iloc[:, loc:]
        return pandas.concat([first_mask, value, second_mask], axis=axis)

    md_frames, pd_frames = create_test_dfs(data, post_fn=post_fn)
    md_item1, md_item2 = md_frames
    pd_item1, pd_item2 = pd_frames

    index_loc = get_loc(pd_item1, loc)

    pd_res = get_reference(pd_item1, loc=index_loc, value=pd_item2)
    md_res = md_item1._query_compiler.insert_item(
        axis=axis, loc=index_loc, value=md_item2._query_compiler, replace=replace
    ).to_pandas()
    # breakpoint()
    df_equals(md_res, pd_res)

    index_loc = get_loc(pd_item2, loc)

    pd_res = get_reference(pd_item2, loc=index_loc, value=pd_item1)
    md_res = md_item2._query_compiler.insert_item(
        axis=axis, loc=index_loc, value=md_item1._query_compiler, replace=replace
    ).to_pandas()

    df_equals(md_res, pd_res)
